from yta_editor_nodes.processor.base import _NodeProcessor
from typing import Union


class SelectionMaskNodeProcessor(_NodeProcessor):
    """
    Class to use a mask selection (from which we will
    determine if the pixel must be applied or not) to
    apply the processed input over the original one.

    If the selection mask is completely full, the
    result will be the processed input.

    Mandatory inputs:
    - `original_input`
    - `processed_input`
    - `selection_mask_input`

    Optional inputs:
    - `None`

    Parameters:
    - `factor`

    The parameters above are the accepted by this
    node, where `*` means that the parameter is
    mandatory.
    """
    
    mandatory_inputs = ['original_input', 'processed_input', 'selection_mask_input']
    optional_inputs = []
    parameters_specifications = []

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        super().__init__(
            opengl_context = opengl_context
        )

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU procesors and return 
        them in that order.

        This method must be implemented by each class.
        """
        from yta_editor_nodes_cpu.processor.selection_mask import SelectionMaskNodeProcessorCPU
        from yta_editor_nodes_gpu.processor.selection_mask import SelectionMaskNodeProcessorGPU

        node_cpu = SelectionMaskNodeProcessorCPU()
        node_gpu = SelectionMaskNodeProcessorGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu


from yta_editor_nodes.processor.base import _ProcessorGPUAndCPU
from typing import Union
from abc import abstractmethod


class _TransitionProcessor(_ProcessorGPUAndCPU):
    """
    *Abstract class*
    
    *Singleton class*

    *For internal use only*

    *This class must be inherited by the specific
    implementation of some transition that will be
    done by CPU or GPU (at least one of the options)*

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.

    Mandatory inputs:
    - `first_input`
    - `second_input`

    Optional inputs:
    - `None`

    Parameters:
    - `None`

    The parameters above are the accepted by this
    node, where `*` means that the parameter is
    mandatory.
    """

    mandatory_inputs = ['first_input', 'second_input']
    optional_inputs = []
    parameters_specifications = []

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None],
        **kwargs
    ):
        """
        The `do_use_gpu` boolean flag will be set by the
        user when instantiating the class to choose between
        GPU and CPU to process the input.
        """
        transition_processor_cpu, transition_processor_gpu = self._instantiate_cpu_and_gpu_processors(
            opengl_context = opengl_context,
            **kwargs
        )

        super().__init__(
            processor_cpu = transition_processor_cpu,
            processor_gpu = transition_processor_gpu,
            opengl_context = opengl_context
        )

    @abstractmethod
    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None],
        **kwargs
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU processors and return 
        them in that order.

        This method must be implemented by each class.
        """
        pass


from yta_editor_nodes.complex.base import _NodeComplex
from yta_editor_parameters.specifications import ParameterSpecification
from typing import Union


class DisplayOverAtNodeComplex(_NodeComplex):
    """
    The `overlay_input` is positioned in the given `position`,
    with the `rotation` and `size` provided, and then put as
    an overlay of the also given `base_input`.

    Information:
    - The scene size is `(1920, 1080)`, so provide the
    `position` parameter according to it. Giving a `(0, 0)`
    means that the center of the `overlay_input` will be in
    that position.
    - The `rotation` is in degrees, where `rotation=90`
    means rotating 90 degrees to the right.
    - The `size` parameter must be provided according to
    the previously mentioned scene size `(1920, 1080)`.

    TODO: This has no inheritance, is special and we need
    to be able to identify it as a valid one.

    Mandatory inputs:
    - `base_input`
    - `overlay_input`

    Optional inputs:
    - `None`

    Parameters:
    - `*x`
    - `*y`
    - `*width`
    - `*height`
    - `*rotation`

    The parameters above are the accepted by this
    node, where `*` means that the parameter is
    mandatory.
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []
    parameters_specifications = [
        # TODO: Check if these limits below are ok
        ParameterSpecification(
            name = 'x',
            type = int,
            is_required = True,
            default = int(1920 / 2),
            min = -9999,
            max = 9999
        ),
        ParameterSpecification(
            name = 'y',
            type = int,
            is_required = True,
            default = int(1080 / 2),
            min = -9999,
            max = 9999
        ),
        ParameterSpecification(
            name = 'width',
            type = int,
            is_required = True,
            default = int(1920 / 2),
            min = -9999,
            max = 9999
        ),
        ParameterSpecification(
            name = 'height',
            type = int,
            is_required = True,
            default = int(1080 / 2),
            min = -9999,
            max = 9999
        ),
        ParameterSpecification(
            name = 'rotation',
            type = int,
            is_required = True,
            default = 45,
            min = -360 * 10,
            max = 360 * 10
        )
    ]

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU blenders and return 
        them in that order.

        This method must be implemented by each class.
        """
        from yta_editor_nodes_cpu.complex.display_over_at import DisplayOverAtNodeComplexCPU
        from yta_editor_nodes_gpu.complex.display_over_at import DisplayOverAtNodeComplexGPU

        node_cpu = DisplayOverAtNodeComplexCPU()
        node_gpu = DisplayOverAtNodeComplexGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu
    
    def _format_parameters(
        self,
        parameters: dict[str, any]
    ) -> dict[str, any]:
        # The 'x' and 'y' exist always
        parameters['position'] = (parameters.pop('x'), parameters.pop('y'))
        # The 'width' and 'height' exist always
        parameters['size'] = (parameters.pop('width'), parameters.pop('height'))
        
        return parameters
    
    """
    I maintain this code below because of the way
    we are handling the optional 'base_input'. I'm
    trying to refactor and simplify everything, but
    I'll keep this code until I confirm this node
    is working perfectly after the refactor.
    """
    # def process(
    #     self,
    #     inputs: Union['np.ndarray', 'moderngl.Texture'],
    #     evaluation_context: EvaluationContext,
    #     output_size: Union[tuple[int, int], None] = None,
    #     do_use_gpu: bool = True,
    #     position: tuple[int, int] = (1920 / 2, 1080 / 2),
    #     size: tuple[int, int] = (1920 / 2, 1080 / 2),
    #     rotation: int = 0
    # ) -> Union['np.ndarray', 'moderngl.Texture']:
    #     """
    #     Put the `overlapy_input` over the `base_input` in
    #     the `position` given and with the `size` and
    #     `rotation` also provided.
    #     """
    #     base_input = inputs['base_input']
    #     overlay_input = inputs['overlay_input']

    #     output_size = (
    #         get_input_size(base_input)
    #         if output_size is None else
    #         output_size
    #     )

    #     processor = self._get_processor(
    #         do_use_gpu = do_use_gpu
    #     )

    #     return processor.process(
    #         inputs = {
    #             'base_input': self._prepare_input(
    #                 input = base_input,
    #                 do_use_gpu = do_use_gpu
    #             ),
    #             'overlay_input': self._prepare_input(
    #                 input = overlay_input,
    #                 do_use_gpu = do_use_gpu
    #             ),
    #         },
    #         evaluation_context = evaluation_context,
    #         output_size = output_size,
    #         position = position,
    #         size = size,
    #         rotation = rotation
    #     )


from yta_editor_nodes.processor.video.transitions.base import _TransitionProcessor
from yta_editor_parameters.specifications import ParameterSpecification
from typing import Union


class BarsFallingTransitionProcessor(_TransitionProcessor):
    """
    A transition between the frames of 2 videos in which
    a set of bars fall with the first video to let the
    second one be seen.

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.

    Mandatory inputs:
    - `first_input`
    - `second_input`

    Optional inputs:
    - `None`
    
    Parameters:
    - `*number_of_bars`
    - `*amplitude`
    - `*noise`
    - `*frequency`
    - `*drip_scale`

    The parameters above are the accepted by this
    node, where `*` means that the parameter is
    mandatory.
    """

    mandatory_inputs = ['first_input', 'second_input']
    optional_inputs = []
    parameters_specifications = [
        ParameterSpecification(
            name = 'number_of_bars',
            type = int,
            is_required = True,
            default = 30,
            min = 5,
            max = 100
        ),
        ParameterSpecification(
            name = 'amplitude',
            type = float,
            is_required = True,
            default = 2.0,
            min = 0.1,
            max = 10.0
        ),
        ParameterSpecification(
            name = 'noise',
            type = float,
            is_required = True,
            default = 0.1,
            min = 0.001,
            max = 1.0
        ),
        ParameterSpecification(
            name = 'frequency',
            type = float,
            is_required = True,
            default = 0.5,
            min = 0.001,
            max = 2.0
        ),
        ParameterSpecification(
            name = 'drip_scale',
            type = float,
            is_required = True,
            default = 0.5,
            min = 0.001,
            max = 5.0
        )
    ]

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        super().__init__(
            opengl_context
        )

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU processors and return 
        them in that order.
        """
        from yta_editor_nodes_gpu.processor.video.transitions.bars_falling import BarsFallingTransitionNodeProcessorGPU

        node_cpu = None
        node_gpu = BarsFallingTransitionNodeProcessorGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu


from yta_editor_nodes.processor.video.transitions.base import _TransitionProcessor
from yta_editor_parameters.specifications import ParameterSpecification
from typing import Union


class CircleOpeningTransitionProcessor(_TransitionProcessor):
    """
    A transition between the frames of 2 videos in
    which the frames are mixed by generating a circle
    that grows from the middle to end fitting the 
    whole screen.

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.

    Mandatory inputs:
    - `first_input`
    - `second_input`

    Optional inputs:
    - `None`
      
    Parameters:
    - `*border_smoothness`

    The parameters above are the accepted by this
    node, where `*` means that the parameter is
    mandatory.
    """

    mandatory_inputs = ['first_input', 'second_input']
    optional_inputs = []
    parameters_specifications = [
        ParameterSpecification(
            name = 'border_smoothness',
            type = float,
            is_required = True,
            default = 0.02,
            min = 0.001,
            max = 1.0
        )
    ]

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        super().__init__(
            opengl_context = opengl_context
        )

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU processors and return 
        them in that order.
        """
        from yta_editor_nodes_gpu.processor.video.transitions.circle_opening import CircleOpeningTransitionNodeProcessorGPU

        node_cpu = None
        node_gpu = CircleOpeningTransitionNodeProcessorGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu


from yta_video_opengl.context.provider import OpenGLContextProvider
from yta_editor_time.evaluation_context import EvaluationContext
from yta_editor_frame.decorators import force_framebuffer
from yta_editor_frame.buffer import FrameBuffer
from yta_editor_nodes_common import _Node
from typing import Union


class _ProcessorGPUAndCPU(_Node):
    """
    *For internal use only*

    *Abstract class*

    Abstract class to share the common behaviour of
    being able to handle a process with a GPU and/or
    a CPU processor, chosen by the user.

    Mandatory inputs:
    - `None`

    Optional inputs:
    - `None`
    """

    mandatory_inputs = []
    optional_inputs = []
    parameters_specifications = []

    @property
    def is_gpu_available(
        self
    ) -> bool:
        """
        Boolean flag to indicate if the GPU is available
        or not, that means that the processor that uses
        GPU is set.
        """
        return self._processor_gpu is not None
    
    @property
    def is_cpu_available(
        self
    ) -> bool:
        """
        Boolean flag to indicate if the CPU is available
        or not, that means that the processor that uses
        CPU is set.
        """
        return self._processor_cpu is not None
    
    def __init__(
        self,
        processor_cpu: Union['_ProcessorGPU', None] = None,
        processor_gpu: Union['_ProcessorCPU', None] = None,
        opengl_context: Union['moderngl.Context', None] = None,
    ):
        """
        The `processor_cpu` and `processor_gpu` have to be
        set by the developer when building the specific
        classes, but the `do_use_gpu` boolean flag will be
        set by the user when instantiating the class to
        choose between GPU and CPU.

        The `opengl_context` is needed to instantiate the
        GPU unit but also to transform the input into a
        `moderngl.Texture` if needed.
        """
        if (
            processor_cpu is None and
            processor_gpu is None
        ):
            raise Exception('No node processor provided. At least one node processor is needed.')

        self._processor_cpu: Union['_ProcessorCPU', None] = processor_cpu
        """
        *For internal use only*

        The transition processor that is able to do the
        processing by using the CPU. If it is None we cannot
        process it with CPU.
        """
        self._processor_gpu: Union['_ProcessorGPU', None] = processor_gpu
        """
        *For internal use only*

        The transition processor that is able to do the
        processing by using the GPU. If it is None we cannot
        process it with GPU.
        """
        self._opengl_context: Union['moderngl.Context', None] = (
            OpenGLContextProvider().context
            if opengl_context is None else
            opengl_context
        )
        """
        *For internal use only*

        The context that we will use for the GPU node.
        """

    def _get_processor(
        self,
        do_use_gpu: bool
    ) -> Union['ProcessorGPU', 'ProcessorCPU']:
        """
        *For internal use only*

        Get the processor according to the `do_use_gpu`
        parameter provided and if the one requested is
        available or not.
        """
        return (
            (
                # Prefer GPU if available
                self._processor_gpu or
                self._processor_cpu
            ) if do_use_gpu else (
                # Prefer CPU if available
                self._processor_cpu or
                self._processor_gpu
            )
        )
    
    def _do_use_gpu(
        self,
        do_use_gpu: bool
    ) -> bool:
        """
        *For internal use only*

        Check if we should use GPU or not, according
        to what the user is requesting with the
        `do_use_gpu` parameter but checking if it is
        available in the node itself.

        This method is useful to know if we have to
        transform a frame to a specific format (for
        GPU or CPU) before sending it as an argument.
        """
        from yta_editor_nodes.utils import _do_use_gpu

        return _do_use_gpu(
            node_processor = self,
            do_use_gpu = do_use_gpu
        )

    # TODO: Maybe move to a more general thing (?)
    def _prepare_input(
        self,
        input: 'FrameBuffer',
        do_use_gpu: bool
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        """
        *For internal use only*

        Prepare the `input` provided based on the `do_use_gpu`
        parameter given to be used with the processer,
        according to the availability.

        This method will transform a `texture` into a `numpy`
        array if they are asking for CPU and it is available,
        or a `numpy` into a `texture` if they are asking for
        GPU and it is available.
        """
        return (
            input.as_moderngl_texture(
                moderngl_context = self._opengl_context
            )
            if self._do_use_gpu(do_use_gpu) else
            input.as_numpy_ndarray()
        )
    
    def _validate_and_prepare_inputs_for_node(
        self,
        inputs: dict[str, 'FrameBuffer'],
        do_use_gpu: bool = True
    ) -> dict[str, Union['np.ndarray', 'moderngl.Texture']]:
        """
        *For internal use only*

        Validate the `inputs` provided and prepare and
        convert them (if needed) to be used in CPU or GPU
        according to the `do_use_gpu` parameter.
        
        This must be done before sending the `inputs` to
        the node processor.
        """
        # 1. Validate
        super()._validate_inputs(inputs, 'FrameBuffer')

        # 2. Prepare for the node (GPU or CPU)
        for input_name in inputs:
            inputs[input_name] = (
                self._prepare_input(
                    input = inputs[input_name],
                    do_use_gpu = do_use_gpu
                )
                if inputs[input_name] is not None else
                None
            )

        return inputs
    
    def _format_parameters(
        self,
        parameters: dict[str, any]
    ) -> dict[str, any]:
        """
        *For internal use only*

        Format the parameters to be sent to the node.
        This method must be called after the parameters
        have been validated and prepared, and before
        sending them to the node so we can format some
        of them.

        For example, we use `x` and `y` parameters that
        we transform into a `position` tuple when being
        sent to the node. This is what must be done here.
        """
        return parameters

    @force_framebuffer('inputs')
    def process(
        self,
        inputs: dict[str, 'FrameBuffer'],
        evaluation_context: EvaluationContext,
        output_size: Union[tuple[int, int], None],
        do_use_gpu: bool = True,
        parameters: dict[str, any] = {}
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        """
        Prepare the `inputs` provided to be processed by
        this processor node according to if the GPU or
        CPU is requested and available, returning the
        result as a `moderngl.Texture` or a `np.ndarray`.
        """
        inputs = self._validate_and_prepare_inputs_for_node(
            inputs = inputs,
            do_use_gpu = do_use_gpu
        )

        first_input = next(iter(inputs.values()))

        # Output size of first input if not forced
        output_size = (
            first_input.size
            if output_size is None else
            output_size
        )

        parameters = self._validate_and_prepare_parameters(
            parameters = parameters
        )
        parameters = self._format_parameters(
            parameters = parameters
        )

        processor = self._get_processor(
            do_use_gpu = do_use_gpu
        )

        return processor.process(
            inputs = inputs,
            evaluation_context = evaluation_context,
            output_size = output_size,
            # TODO: Should we pass 'dict' and parse again instead (?)
            **parameters
        )


from yta_editor_nodes.blender.base import _Blender
from yta_editor_parameters.specifications import ParameterSpecification
from typing import Union


class MixBlender(_Blender):
    """
    A blender that uses a float value to mix the
    result with the original input as much as that
    value determines.

    Mandatory inputs:
    - `base_input`
    - `overlay_input`

    Optional inputs:
    - `None`

    Parameters:
    - `*mix_weight`

    The parameters above are the accepted by this
    node, where `*` means that the parameter is
    mandatory.
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []
    parameters_specifications = [
        ParameterSpecification(
            name = 'mix_weight',
            type = float,
            is_required = True,
            default = 1.0,
            min = 0.01,
            max = 1.0
        )
    ]

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        super().__init__(
            opengl_context = opengl_context
        )

    def _instantiate_cpu_and_gpu_blenders(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU blenders and return 
        them in that order.

        This method has been created to be used in both the
        '__init__' and the '__reinit__' methods.

        This method must be implemented by each class.
        """
        from yta_editor_nodes_cpu.blender.mix import MixBlenderCPU
        from yta_editor_nodes_gpu.blender.mix import MixBlenderGPU

        node_cpu = MixBlenderCPU()
        node_gpu = MixBlenderGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu


from yta_editor_nodes.processor.video.base import _VideoNodeProcessor
from typing import Union


class AlphaPediaMaskRevealVideoNodeProcessor(_VideoNodeProcessor):
    """
    A transition made by using a custom mask to
    reveal a video. This mask is specifically obtained
    from the AlphaPediaYT channel in which we upload
    specific masking videos.

    The video will be placed ocuppying the whole scene
    and revealed progressively by using the video mask.

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.

    Mandatory inputs:
    - `base_input`
    - `mask_input`

    Optional inputs:
    - `None`

    Parameters:
    - `None`

    The parameters above are the accepted by this
    node, where `*` means that the parameter is
    mandatory.
    """

    mandatory_inputs = ['base_input', 'mask_input']
    optional_inputs = []
    parameters_specifications = []

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU processors and return 
        them in that order.
        """
        # TODO: Create the processor for CPU if possible
        from yta_editor_nodes_gpu.processor.video.alphapedia_mask_reveal import AlphaPediaMaskRevealVideoNodeProcessorGPU

        node_cpu = None
        node_gpu = AlphaPediaMaskRevealVideoNodeProcessorGPU(
            opengl_context = opengl_context,
        )

        return node_cpu, node_gpu


from yta_editor_nodes.blender.base import _Blender
from yta_editor_parameters.specifications import ParameterSpecification
from typing import Union


class AddBlender(_Blender):
    """
    The most common blender used in video edition.

    This blender will increase the brightness by
    combining the colors of the base and the overlay
    inputs, using the overlay as much as the 
    `strength` parameter is indicating.

    Mandatory inputs:
    - `base_input`
    - `overlay_input`

    Optional inputs:
    - `None`

    Parameters:
    - `*mix_weight`
    - `*strength`

    The parameters above are the accepted by this
    node, where `*` means that the parameter is
    mandatory.
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []
    parameters_specifications = [
        ParameterSpecification(
            name = 'mix_weight',
            type = float,
            is_required = True,
            default = 1.0,
            min = 0.01,
            max = 1.0
        ),
        ParameterSpecification(
            name = 'strength',
            type = float,
            is_required = True,
            default = 1.0,
            min = 0.01,
            max = 1.0
        )
    ]

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None],
    ):
        super().__init__(
            opengl_context = opengl_context
        )

    def _instantiate_cpu_and_gpu_blenders(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU blenders and return 
        them in that order.

        This method must be implemented by each class.
        """
        from yta_editor_nodes_cpu.blender.add import AddBlenderCPU
        from yta_editor_nodes_gpu.blender.add import AddBlenderGPU

        node_cpu = AddBlenderCPU()
        node_gpu = AddBlenderGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu


from yta_editor_nodes.processor.base import _NodeProcessor
from yta_editor_parameters.specifications import ParameterSpecification
from typing import Union


class ColorContrastNodeProcessor(_NodeProcessor):
    """
    The node to modify the color contrast of an input
    by using the CPU or the GPU.

    Mandatory inputs:
    - `base_input`

    Optional inputs:
    - `None`

    Parameters:
    - `factor`

    The parameters above are the accepted by this
    node, where `*` means that the parameter is
    mandatory.
    """

    mandatory_inputs = ['base_input']
    optional_inputs = []
    parameters_specifications = [
        ParameterSpecification(
            name = 'factor',
            type = float,
            is_required = True,
            default = 1.0,
            min = 0.0,
            max = 5.0
        )
    ]

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        super().__init__(
            opengl_context = opengl_context
        )

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU procesors and return 
        them in that order.

        This method must be implemented by each class.
        """
        from yta_editor_nodes_cpu.processor.color_contrast import ColorContrastNodeProcessorCPU
        from yta_editor_nodes_gpu.processor.color_contrast import ColorContrastNodeProcessorGPU

        node_cpu = ColorContrastNodeProcessorCPU()
        node_gpu = ColorContrastNodeProcessorGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu


def _do_use_gpu(
    # TODO: Set type
    node_processor: any,
    do_use_gpu: bool = True
) -> bool:
    """
    *For internal use only*

    Check if we should use GPU or not, according
    to what the user is requesting with the
    `do_use_gpu` parameter but checking if it is
    available in the `node_processor` or not.

    This method is useful to know if we have to
    transform a frame to a specific format (for
    GPU or CPU) before sending it as an argument.
    """
    # TODO: Validate 'node_processor'
    return (
        (
            do_use_gpu and
            node_processor.is_gpu_available
        )
    ) or (
        (
            not do_use_gpu and
            not node_processor.is_cpu_available
        )
    )


from yta_editor_nodes.timeline import TimelineNode
from yta_editor_nodes.timeline.serial import SerialTimelineNode
from yta_editor_time.evaluation_context import EvaluationContext
from abc import ABC, abstractmethod
from typing import Union


class _ParallelTimelineNodeAbstract(TimelineNode, ABC):
    """
    The abstract class of the parallel node.

    This class is limited and needs at least 2 nodes
    to be valid.
    """

    def __init__(
        self,
        # TODO: Put the correct class
        # TODO: Maybe I need the 'node' in the TimelineNode class
        # TODO: Maybe I need a shortcut to 'is_gpu_available'...
        name: str,
        nodes: list[SerialTimelineNode]
    ):
        if len(nodes) < 2:
            raise Exception('A ParallelTimelineNode needs 2 or more nodes to be created.')
        
        self.nodes: list[SerialTimelineNode] = nodes
        """
        The list of nodes that will be executed in parallel
        to obtain the outputs that will be combined.
        """

        super().__init__(
            name = name
        )

    def _resolve_specification(
        self,
        evaluation_context: EvaluationContext
    ) -> 'TimelineNode':
        """
        *For internal use only*

        Resolve the specification of this node by using the
        `evaluation_context` provided, to be able to know
        when the node is active and when we should process
        the input.

        This method must be called each time we modify
        something that affects to the `duration` or other
        field that makes the specifications resolved be
        different.
        """
        for node in self.nodes:
            node._resolve_specification(evaluation_context)

        return self

    # TODO: I think I need an 'is_active_at' method
    def is_active_at(
        self,
        evaluation_context: EvaluationContext
    ) -> bool:
        """
        Flag to indicate if the `evaluation_context`
        provided is in the range of this TimedNode instance, 
        which means that it is in between the limits of this
        `TimelineNode` instance.

        The formula:
        - `start_frame <= evaluation_context.frame_index < end_frame`
        """
        if not self.is_enabled:
            return False
        
        for node in self.nodes:
            if node.is_active_at(evaluation_context):
                return True
            
        return False

    def process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        evaluation_context: EvaluationContext,
        output_size: tuple[int, int],
        do_use_gpu: bool = True,
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the `inputs` provided for the
        `evaluation_context` given, and combine the
        outputs of the different internal nodes and
        return them as output.
        """
        active_nodes = [
            node for node in self.nodes
            if node.is_active_at(evaluation_context)
        ]

        # This shouldn't happen as we know that the parallel
        # node itself is not active, but just in case...
        if not active_nodes:
            # TODO: What do we do with the different inputs? We
            # should return only one from the dict...
            return inputs

        if len(active_nodes) == 1:
            return active_nodes[0].process(
                inputs = inputs,
                evaluation_context = evaluation_context,
                output_size = output_size,
                do_use_gpu = do_use_gpu,
                **kwargs
            )

        outputs = [
            node.process(
                inputs = inputs,
                evaluation_context = evaluation_context,
                output_size = output_size,
                do_use_gpu = do_use_gpu,
                **kwargs
            )
            for node in active_nodes
        ]

        return self._combine_outputs(outputs)
    
    @abstractmethod
    def _combine_outputs(
        self,
        outputs: list[Union['np.ndarray', 'moderngl.Texture']]
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Combine the different `outputs` into a single one.

        This class must be overwritten by every specific
        class implementation.
        """
        # TODO: This must be according to the type
        pass


from yta_editor_nodes.processor.base import _ProcessorGPUAndCPU
from typing import Union
from abc import abstractmethod


class _NodeCompositor(_ProcessorGPUAndCPU):
    """
    *For internal use only*

    This class must be inherited by the specific
    implementation of some node that will be positioning
    inputs, done by CPU or GPU (at least one of the
    options)

    A node specifically designed to build a scene by
    positioning inputs in different positions and 
    obtaining a single output by using GPU or CPU.

    Mandatory inputs:
    - `base_input`

    Optional inputs:
    - `None`

    Parameters:
    - `None`

    The parameters above are the accepted by this
    node, where `*` means that the parameter is
    mandatory.
    """

    mandatory_inputs = ['base_input']
    optional_inputs = []
    parameters_specifications = []

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None],
        **kwargs
    ):
        node_compositor_cpu, node_compositor_gpu = self._instantiate_cpu_and_gpu_processors(
            opengl_context = opengl_context,
            **kwargs
        )

        super().__init__(
            processor_cpu = node_compositor_cpu,
            processor_gpu = node_compositor_gpu,
            opengl_context = opengl_context
        )

    @abstractmethod
    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None],
        **kwargs
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU procesors and return 
        them in that order.

        This method must be implemented by each class.
        """
        pass


from yta_editor_nodes.timeline.layer.abstract import _LayerNodeAbstract
from abc import abstractmethod
from typing import Union


class LayerNode(_LayerNodeAbstract):
    """
    A node that is built by layers whose outputs are
    combined into a single one.

    TODO: I don't really know how this works...

    TODO: This is just a basic class using the MixBlender
    to combine the outputs. We can create more...
    """

    @abstractmethod
    def _blend(
        self,
        base: Union['np.ndarray', 'moderngl.Texture'],
        top: Union['np.ndarray', 'moderngl.Texture'],
        # TODO: What is the mode (?)
        mode = str
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Blend the `base` and `top` inputs provided, using
        the also given `mode`, to obtain an output.
        """
        # TODO: This must be according to the type
        from yta_editor_nodes.blender import MixBlender

        blender = MixBlender(
            do_use_gpu = True
        )

        return blender.process(
            inputs = {
                'base_input': base,
                'overlay_input': top
            },
            mix_weight = 1.0
        )


from yta_editor_nodes.processor.video.transitions.base import _TransitionProcessor
from typing import Union


class SlideTransitionProcessor(_TransitionProcessor):
    """
    A transition in which the frames goes from one
    side to the other, disappearing the first one
    and appearing the second one.

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.

    Mandatory inputs:
    - `first_input`
    - `second_input`

    Optional inputs:
    - `None`
    
    Parameters:
    - `None`

    The parameters above are the accepted by this
    node, where `*` means that the parameter is
    mandatory.
    """

    mandatory_inputs = ['first_input', 'second_input']
    optional_inputs = []
    parameters_specifications = []

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None],
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU processors and return 
        them in that order.
        """
        # TODO: Maybe rename to 'SlideTransitionNodeProcessorGPU' (?)
        from yta_editor_nodes_cpu.processor.video.transitions.slide import SlideTransitionProcessorCPU
        from yta_editor_nodes_gpu.processor.video.transitions.slide import SlideTransitionNodeProcessorGPU

        node_cpu = SlideTransitionProcessorCPU()
        node_gpu = SlideTransitionNodeProcessorGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu


"""
The blenders, capable of mixing different inputs into
a single one by using different parameters and 
techniques.

All the classes here will have an instance of the
specific CPU and/or GPU class that is able to run the
code by using either CPU or GPU. The user can choose
between GPU and CPU and that option will be considered
(only if available).
"""
from yta_editor_nodes.blender.mix import MixBlender
from yta_editor_nodes.blender.alpha import AlphaBlender
from yta_editor_nodes.blender.add import AddBlender


__all__ = [
    'MixBlender',
    'AlphaBlender',
    'AddBlender'
]


"""
Module for the nodes that can be applied to
the timeline as edition nodes. These are the
only ones that can be added, and will include
the processing nodes that will make the 
operations.
"""
from yta_editor_nodes.timeline.abstract import TimelineNode
from typing import Union


# TODO: Check the 'new_nodes.py' module for other
# things we should maybe include
# TODO: Other nodes that we are not, by now, handling
"""
This node is to apply changes on the inverse of
the mask of the previous input. TODO: I need to
handle the masks to be able to do this...
"""
# class OutsideNode(TimelineNode):
#     def __init__(self, inside_node: TimelineNode, mask_func):
#         self.inside_node = inside_node
#         self.mask_func = mask_func  # callable: input -> mask (0/1)

#     def process(self, input, **kwargs):
#         mask = self.mask_func(input)
#         inside = self.inside_node.process(input, **kwargs)
#         return input * mask + inside * (1 - mask)

"""
TODO: I don't know the utility of this one...
"""
# class SplitterCombinerNode(TimelineNode):
#     def __init__(self, channels=('R', 'G', 'B')):
#         self.channels = channels

#     def process(self, input, **kwargs):
#         splits = [self._extract_channel(input, c) for c in self.channels]
#         return self._combine_channels(splits)

#     def _extract_channel(self, img, ch):
#         idx = {'R': 0, 'G': 1, 'B': 2}[ch]
#         return img[..., idx:idx+1]

#     def _combine_channels(self, splits):
#         return np.concatenate(splits, axis=-1)

"""
A node that just makes different changes in
serie for the same input.
"""
class CompoundNode(TimelineNode):
    """
    A node that has different nodes that are applied
    to the same input, one after the other, to return
    a single output.

    It is equivalent to use different SerialNode nodes
    in sequence.
    """

    def __init__(
        self,
        nodes: list[TimelineNode]
    ):
        self.nodes: list[TimelineNode] = nodes
        """
        The nodes to execute one after another to obtain 
        the output.
        """

    def process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        # TODO: Maybe 'evaluation_context' (?)
        # TODO: Maybe 'output_size' (?)
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the input and obtain a result as the output.
        """
        result = input

        for node in self.nodes:
            result = node.process(
                inputs = inputs,
                **kwargs
            )

        return result
    
class SharedNode(TimelineNode):
    """
    A node that is shared in between different clips to
    apply the same modification.

    TODO: I'm not sure if I will use this one...
    """

    def __init__(
        self,
        node: TimelineNode,
        cached_output: Union['np.ndarray', 'moderngl.Texture', None]
    ):
        self.node: TimelineNode = node
        """
        The node to execute and to obtain the output from.
        """
        self.cached_output: Union['np.ndarray', 'moderngl.Texture', None] = cached_output
        """
        The output that is cached in this instance to be
        able to access to it from other clips and/or nodes
        in the system.
        """

    def process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        # TODO: Maybe 'evaluation_context' (?)
        # TODO: Maybe 'output_size' (?)
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the input and obtain a result as the output,
        that will be cached in this instance to be able to
        access to it, or returned if it's been previously
        applied.
        """
        if self.cached_output is None:
            # TODO: Is this useful if we only calculate it
            # once (?)
            self.cached_output = self.node.process(
                inputs = inputs,
                **kwargs
            )

        return self.cached_output
    
"""
TODO: I don't understand this node, but it was
suggested so maybe it is useful...
"""
# class FloatingNode(TimelineNode):
#     """
#     TODO: A node that I don't understand at all...
#     """
#     def __init__(self, node: TimelineNode, blend_mode="add"):
#         self.node = node
#         self.blend_mode = blend_mode

#     def process(self, input, **kwargs):
#         base = input
#         float_out = self.node.process(input, **kwargs)
#         return self._blend(base, float_out, self.blend_mode)

#     def _blend(self, base, top, mode):
#         if mode == "add":
#             return base + top
#         elif mode == "multiply":
#             return base * top
#         else:
#             return 0.5 * base + 0.5 * top


from yta_editor_nodes.processor.base import _ProcessorGPUAndCPU
from yta_editor_parameters.specifications import ParameterSpecification
from typing import Union
from abc import abstractmethod


class _Blender(_ProcessorGPUAndCPU):
    """
    *For internal use only*

    This class must be inherited by the specific
    implementation of some blenders that will use
    CPU or GPU (at least one of the options).

    Class that is capable of mixing 2 different
    inputs by using the GPU or the CPU.

    Mandatory inputs:
    - `base_input`
    - `overlay_input`

    Optional inputs:
    - `None`

    Parameters:
    - `*mix_weight`

    The parameters above are the accepted by this
    node, where `*` means that the parameter is
    mandatory.
    """

    mandatory_inputs = ['base_input', 'overlay_input']
    optional_inputs = []
    parameters_specifications = [
        ParameterSpecification(
            name = 'mix_weight',
            type = float,
            is_required = True,
            default = 1.0,
            min = 0.01,
            max = 1.0
        )
    ]

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None] = None,
        **kwargs
    ):
        """
        The `do_use_gpu` boolean flag will be set by the
        user when instantiating the class to choose between
        GPU and CPU to process the input.
        """
        blender_cpu, blender_gpu = self._instantiate_cpu_and_gpu_blenders(
            opengl_context = opengl_context,
            **kwargs
        )

        super().__init__(
            processor_cpu = blender_cpu,
            processor_gpu = blender_gpu,
            opengl_context = opengl_context
        )

    @abstractmethod
    def _instantiate_cpu_and_gpu_blenders(
        self,
        opengl_context: Union['moderngl.Context', None],
        **kwargs
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU blenders and return 
        them in that order.

        This method must be implemented by each class.
        """
        pass

    """
    I maintain this code below because of the way
    we are handling the optional 'base_input'. I'm
    trying to refactor and simplify everything, but
    I'll keep this code until I confirm this node
    is working perfectly after the refactor.
    """
    # def process(
    #     self,
    #     inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
    #     # TODO: Maybe set it automatically based on the texture
    #     # size (?)
    #     evaluation_context: EvaluationContext,
    #     output_size: Union[tuple[int, int], None] = None,
    #     do_use_gpu: bool = True,
    #     mix_weight: float = 1.0,
    #     **kwargs
    # # TODO: What about the output type (?)
    # ) -> Union['np.ndarray', 'moderngl.Texture']:
    #     """
    #     Process the provided 'base_input' and 'overlay_input'
    #     with GPU or CPU according to the internal flag.

    #     The `mix_weight` will determine how much of the result
    #     we keep against the base, where a `mix_weight == 1.0`
    #     means that we want the result at 100%, but a
    #     `mix_weight == 0.4` means a 40% of the result and a 60%
    #     of the base.
    #     """
    #     self._validate_inputs(inputs)

    #     base_input = inputs['base_input']
    #     overlay_input = inputs['overlay_input']

    #     output_size = (
    #         get_input_size(base_input)
    #         if output_size is None else
    #         output_size
    #     )

    #     processor = self._get_processor(
    #         do_use_gpu = do_use_gpu
    #     )

    #     # TODO: Transform the inputs if needed, based on
    #     # the type and the processor we got: numpy to
    #     # texture if GPU, texture to numpy if CPU

    #     return processor.process(
    #         inputs = {
    #             'base_input': self._prepare_input(
    #                 input = base_input,
    #                 do_use_gpu = do_use_gpu
    #             ),
    #             'overlay_input': self._prepare_input(
    #                 input = overlay_input,
    #                 do_use_gpu = do_use_gpu
    #             ),
    #         },
    #         evaluation_context = evaluation_context,
    #         output_size = output_size,
    #         mix_weight = mix_weight,
    #         **kwargs
    #     )


from yta_editor_time.evaluation_context import EvaluationContext
from yta_validation.parameter import ParameterValidator
from abc import ABC, abstractmethod
from typing import Union


class TimelineNode(ABC):
    """
    *Abstract class*

    *This class has to be inherited by any class that is
    able to handle some input to obtain an output as a
    result*

    The abstract class of a TimelineNode, which is the
    entity that is able to process some input to return
    an output that can be sent to the next node, and able
    to connect (by storing the references) to the other
    nodes.
    """

    @property
    def has_output_node(
        self
    ) -> bool:
        """
        Boolean flag to indicate if this node has an
        output node or not.
        """
        return self.output_node is not None
    
    @property
    def has_input_node(
        self
    ) -> bool:
        """
        Boolean flag to indicate if this node has an
        input node or not.
        """
        return self.input_node is not None
    
    @property
    def duration(
        self
    ) -> float:
        """
        The duration of the timeline node, in frames. This
        can be calculated only if the specification has been
        resolved.

        The formula:
        - `self._specification_resolved.end_frame - self._specification_resolved.start_frame`
        """
        return self._specification_resolved.end_frame - self._specification_resolved.start_frame

    def __init__(
        self,
        name: str
    ):
        ParameterValidator.validate_mandatory_string('name', name, do_accept_empty = False)

        self.name: str = name
        """
        Just a simple name for the node.
        """
        self.is_enabled: bool = True
        """
        Boolean flag to indicate if the node is enabled or not,
        which means that it will not be working if not enabled.
        Basically, if it is on or off.
        """

        self.input_node: Union['TimelineNode', None] = None
        """
        The node that is connected to this one as an
        input, so the output of that previous node will
        be passed as the input for this one.
        """
        self.output_node: Union['TimelineNode', None] = None
        """
        This node is connected as the input of another
        one that is after this. The output of this one
        will be sent as the input for that one.
        """
        self._cached_output: Union['np.ndarray', 'moderngl.Texture', None] = None
        """
        The output, but cached, so when it is generated
        it is stored here.
        """

    def enable(
        self
    ) -> 'TimelineNode':
        """
        Enable the node by setting the internal flag to
        `True`, which means that will be working.
        """
        self.is_enabled = True

        return self
    
    def disable(
        self
    ) -> 'Timeline':
        """
        Disable the node by setting the internal flag
        to `False`, which means that will be working not.
        """
        self.is_enabled = False

        return self
    
    @abstractmethod
    def _resolve_specification(
        self,
        evaluation_context: EvaluationContext
    ) -> 'TimelineNode':
        """
        *For internal use only*

        Resolve the specification of this node by using the
        `evaluation_context` provided, to be able to know
        when the node is active and when we should process
        the input.

        This method must be called each time we modify
        something that affects to the `duration` or other
        field that makes the specifications resolved be
        different.
        """
        pass
    
    @abstractmethod
    def is_active_at(
        self,
        evaluation_context: EvaluationContext
    ) -> bool:
        """
        Check if any internal node is active and needs
        to be applied for the `evaluation_context`
        provided.
        """
        pass

    def connect_to(
        self,
        node: 'TimelineNode'
    ) -> 'TimelineNode':
        """
        Connect the `node` provided to this one as an
        output, and also this one as an input of the
        `node` provided.

        TODO: The connection has to be done with another
        class that inherits from this 'TimelineNode'.
        """
        self.output_node = node
        node.input_node = self

        return self

    def clear_cache(
        self
    ) -> 'TimelineNode':
        """
        Clear the cache of this node, but also for
        the nodes that are outputs of this one, as
        the results would change.
        """
        # TODO: I think this cache is not valid because
        # we are using a `t` and different inputs...
        self._cached_output = None

        self.output_node.clear_cache()

        return self
    
    @abstractmethod
    def process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        evaluation_context: 'EvaluationContext',
        output_size: tuple[int, int],
        do_use_gpu: bool = True,
        # TODO: I think this has to be gone
        **kwargs
    ) -> Union['moderngl.Texture', 'np.ndarray']:
        """
        Process the `inputs` by using the `t_timeline` time
        moment provided, that could be necessary for the
        node if it is a dynamic one that depends on this
        value.
        """
        pass


"""
Our awesome editor module in which we have
all the classes that interact with it and
make it possible.

This is the nodes module, in which we have
all the classes that make the concept work.
"""
from yta_editor_nodes.timeline.serial import SerialTimelineNode
from yta_editor_nodes.timeline.parallel import ParallelTimelineNode
from yta_editor_frame.buffer import FrameBuffer
from yta_editor_frame.decorators import force_framebuffer
from yta_editor_time.evaluation_context import EvaluationContext
from yta_validation import PythonValidator
from typing import Union


class TimelineGraph:
    """
    Graph of nodes (serie and parallel) to process
    the input, handle it and generate an output.

    The graph includes a starting (root) node that
    will begin the process, and maybe other nodes
    connected to it as output nodes.

    The nodes will be interconnected, having a
    `t_start` and `t_end`. The sequence of nodes
    determines the order, and the time range if
    they should modify the input or not.
    """

    def __init__(
        self,
        # TODO: Know 'timeline' associated to know fps (?)
        root_node: Union[SerialTimelineNode, ParallelTimelineNode]
    ):
        self.root_node: Union[SerialTimelineNode, ParallelTimelineNode] = root_node
        """
        The root node in which everything starts.
        """
        self._last_node: Union[SerialTimelineNode, ParallelTimelineNode] = root_node
        """
        *For internal use only*

        The last node we have and the one that will return
        the output.
        """
        # TODO: This is meant to be cached and not re-resolve
        # the specifications if it is not needed, but idk why
        # it is not working
        # self._last_context_signature: Union[tuple, None] = None
        # """
        # *For internal use only*

        # A tuple including the `fps` and `total_frames` of
        # the las `evaluation_context` that was used to 
        # resolve the specifications of the nodes.

        # This will force to re-resolve the specifications
        # if it is `None` or some of the parameters in it
        # are different than this one.
        # """

    def _connect_node_to_last(
        self,
        node: Union[SerialTimelineNode, ParallelTimelineNode]
    ) -> 'TimelineGraph':
        """
        *For internal use only*

        Connect the `node` provided to the last node of this
        timeline, and set this new node as the last one.
        """
        self._last_node.connect_to(node)
        self._last_node = node

        return self

    def add_node(
        self,
        node: Union[SerialTimelineNode, ParallelTimelineNode]
    ) -> 'TimelineGraph':
        """
        Add the `node` provided to the nodes list, connecting
        it (in serie mode) to the last one.
        """
        # TODO: Validate node (?)
        return self._connect_node_to_last(node)

    def add_parallel_node(
        self,
        nodes: list[SerialTimelineNode]
    ):
        """
        Create a `ParallelTimelineNode` with the `nodes` provided
        as parameter and connect it (in serie mode) to the last
        one.
        """
        parallel_node = ParallelTimelineNode(
            name = 'invented',
            # TODO: Validate nodes (?)
            nodes = nodes
        )
        
        return self._connect_node_to_last(parallel_node)
    
    def _resolve_specifications(
        self,
        evaluation_context: EvaluationContext
    ) -> 'TimelineGraph':
        """
        *For internal use only*

        Resolve all the specifications of the nodes that are
        attached to this `TimelineGraph` instance, to keep
        the specifications resolved in each of those nodes to
        be able to evaluate and render when active.

        This method must be called each time we modify
        something that affects to the `duration` or other
        field that makes the specifications resolved be
        different, by checking the internal variable
        `self._last_evaluation_context_used_to_resolve`.
        """
        node = self.root_node

        while node is not None:
            node._resolve_specification(evaluation_context)
            node = node.output_node

        return self

    # TODO: Maybe rename to 'process' (?)
    @force_framebuffer('inputs')
    def render(
        self,
        inputs: dict[str, Union['FrameBuffer', None]],
        evaluation_context: EvaluationContext,
        output_size: tuple[int, int],
        do_use_gpu: bool = True
    ) -> 'FrameBuffer':
        """
        Render the provided `inputs` with the given
        `evaluation_context` to generate an output according
        to the changes that the nodes that were active made.

        The different nodes of this instance will be applied
        in order and only if they are enabled and active for
        that `evaluation_context`.
        """
        # TODO: I don't know exactly why this is not working
        # current_signature = (
        #     evaluation_context.fps,
        #     evaluation_context.total_frames
        # )

        # do_resolve_specifications = False
        # if self._last_context_signature != current_signature:
        #     do_resolve_specifications = True
        #     self._last_context_signature = current_signature

        do_resolve_specifications = True

        # TODO: Maybe a copy of inputs to not affect by
        # reference (?)
        # We obtain the first available input
        output = next((input for input in inputs.values() if input is not None), None)
        node = self.root_node

        if node is None:
            return output

        while node is not None:
            # TODO: Check if we need to do this always or not
            if do_resolve_specifications:
                node._resolve_specification(evaluation_context)

            if node.is_active_at(
                evaluation_context = evaluation_context
            ):
                output = node.process(
                    inputs = inputs,
                    evaluation_context = evaluation_context,
                    output_size = output_size,
                    do_use_gpu = do_use_gpu
                )

            node = node.output_node

        # We return it as a 'FrameBuffer'
        return FrameBuffer.from_frame(output)
    
__all__ = [
    'TimelineGraph',
    'SerialTimelineNode',
    'ParallelTimelineNode'
]
    

"""
Note for the developer:

A guide to the different types of nodes we will have
when imitating DaVinci Resolve.

TimelineNode (abstract)
├── ProcessorNode (process one single flow)
│   ├── EffectNode (filtros, LUTs, shaders…)
│   └── TransitionNode (mix 2 inputs)
│
├── CompositeNode (combine multiple flows)
│   ├── SerialTimelineNode (one after another)
│   ├── ParallelTimelineNode (different branches simultaneously)
│   └── LayerNode (layers with blending or masks)
│
└── GroupNode (has a subnetwork of nodes)

"""



from yta_editor_nodes.timeline import TimelineNode
from yta_editor_parameters.values.abstract import VideoEditorParameterValue
from yta_editor_parameters.values.utils import parse_parameters
from yta_editor_frame.decorators import force_framebuffer
from yta_editor_nodes.timeline.utils import evaluate_parameters
from yta_editor_time.specifications import FrameIndexSpecification, ProgressSpecification, TimeSpecification
from yta_editor_time.specifications.resolved import FrameIndexSpecificationResolved
from yta_editor_time.specifications.resolver import SpecificationResolver
from yta_editor_time.evaluation_context import EvaluationContext
from typing import Union


class SerialTimelineNode(TimelineNode):
    """
    A node that is executed from a specific input and
    generates a single output that is sent to the next
    node.

    Check the `.process` method definition of the node
    you add to see the parameters you need to execute
    it and pass them through the `parameters`
    parameter.
    """

    @property
    def is_gpu_available(
        self
    ) -> bool:
        """
        Boolean flag to indicate if the GPU is available
        in the node or not.
        """
        return self.node.is_gpu_available

    def __init__(
        self,
        # TODO: Put the correct class
        # TODO: Maybe I need a shortcut to 'is_gpu_available'...
        node: 'NodeProcessor',
        name: str,
        specification: Union[FrameIndexSpecification, ProgressSpecification, TimeSpecification],
        parameters: dict[str, VideoEditorParameterValue] = {}
    ):
        # ParameterValidator.validate_mandatory_dict_of('parameters', parameters, VideoEditorParameterValue)
        # Validate the 'parameters' by accepting
        # basic and non iterable types. This will raise
        # exception if not valid
        parameters = parse_parameters(parameters)

        self.node: 'NodeProcessor' = node
        """
        The node to execute and to obtain the output from.
        """
        self.parameters: dict[str, VideoEditorParameterValue] = parameters
        """
        A dict including the `key` of the parameter and
        its value to be calculated.
        """
        self.specification: Union[FrameIndexSpecification, ProgressSpecification, TimeSpecification] = specification
        """
        The specification of when we should apply the settings,
        that can be based on frame index, time or normalized
        progress.
        """
        self._specification_resolved: Union[FrameIndexSpecificationResolved, None] = None
        """
        *For internal use only*

        The specification that has been resolved according to
        an specific evaluation context. This is the one that
        has to be used to check if the node is active and the
        input provided has to be processed.
        """

        super().__init__(
            name = name
        )

    def _resolve_specification(
        self,
        evaluation_context: EvaluationContext
    ) -> 'TimelineNode':
        """
        *For internal use only*

        Resolve the specification of this node by using the
        `evaluation_context` provided, to be able to know
        when the node is active and when we should process
        the input.

        This method must be called each time we modify
        something that affects to the `duration` or other
        field that makes the specifications resolved be
        different.
        """
        self._specification_resolved = SpecificationResolver.resolve(
            specification = self.specification,
            fps = evaluation_context.fps,
            duration = evaluation_context.duration,
            total_frames = evaluation_context.total_frames
        )

        return self
    
    def is_active_at(
        self,
        evaluation_context: EvaluationContext
    ) -> bool:
        """
        Flag to indicate if the `evaluation_context`
        provided is in the range of this TimedNode instance, 
        which means that it is in between the limits of this
        `TimelineNode` instance.

        The formula:
        - `start_frame <= evaluation_context.frame_index < end_frame`
        """
        return (
            self.is_enabled and
            self._specification_resolved.start_frame <= evaluation_context.frame_index < self._specification_resolved.end_frame
        )

    @force_framebuffer('inputs')
    def process(
        self,
        inputs: dict[str, 'FrameBuffer'],
        evaluation_context: 'EvaluationContext',
        output_size: tuple[int, int],
        do_use_gpu: bool = True,
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the `inputs` provided and obtain a result
        as the output by using the `evaluation_context`
        given.
        """
        # This method is only called if it is active
        self._resolve_specification(evaluation_context)

        # Create a new 'EvaluationContext' to propagate
        evaluation_context = EvaluationContext(
            fps = evaluation_context.fps,
            # TODO: Is this 'total_frames' ok? Does it  affect (?)
            total_frames = evaluation_context.total_frames,
            frame_index = evaluation_context.frame_index - self._specification_resolved.start_frame
        )

        # TODO: Do we evaluate with the new one or with the old?
        evaluated_parameters = evaluate_parameters(
            parameters = self.parameters,
            evaluation_context = evaluation_context
        )

        # Combine but giving priority to **kwargs
        evaluated_parameters = {**evaluated_parameters, **kwargs}

        return self.node.process(
            inputs = inputs,
            evaluation_context = evaluation_context,
            output_size = output_size,
            do_use_gpu = do_use_gpu,
            parameters = evaluated_parameters
        )


from yta_editor_nodes.processor.video.transitions.base import _TransitionProcessor
from yta_editor_parameters.specifications import ParameterSpecification
from typing import Union


class DistortedCrossfadeTransitionProcessor(_TransitionProcessor):
    """
    A transition between the frames of 2 videos,
    transforming the first one into the second one
    with a distortion in between.

    Class that is capable of doing some processing on
    an input by using the GPU or the CPU.

    Mandatory inputs:
    - `first_input`
    - `second_input`

    Optional inputs:
    - `None`

    Parameters:
    - `*intensity`

    The parameters above are the accepted by this
    node, where `*` means that the parameter is
    mandatory.
    """

    mandatory_inputs = ['first_input', 'second_input']
    optional_inputs = []
    parameters_specifications = [
        ParameterSpecification(
            name = 'intensity',
            type = float,
            is_required = True,
            default = 1.0,
            min = 0.1,
            max = 10.0
        )
    ]

    def _instantiate_cpu_and_gpu_processors(
        self,
        opengl_context: Union['moderngl.Context', None]
    ):
        """
        *For internal use only*

        Instantiate the CPU and GPU processors and return 
        them in that order.
        """
        from yta_editor_nodes_gpu.processor.video.transitions.distorted_crossfade import DistortedCrossfadeTransitionNodeProcessorGPU

        node_cpu = None
        node_gpu = DistortedCrossfadeTransitionNodeProcessorGPU(
            opengl_context = opengl_context
        )

        return node_cpu, node_gpu


from yta_editor_nodes.timeline import TimelineNode
from abc import ABC, abstractmethod
from typing import Union


# TODO: This needs to be refactored like the
# 'parallel' and 'serie' with new parameters
# and structure
class _LayerNodeAbstract(TimelineNode, ABC):
    """
    The abstract class of the parallel node.
    """

    def __init__(
        self,
        layers: list[TimelineNode],
        blend_modes: list[str]
    ):
        self.layers: list[TimelineNode] = layers
        """
        The list of layers that will be combined to generate
        a single output.
        """
        self.blend_modes: list[str] = blend_modes
        # TODO: What about the blend mode (?)
        # TODO: Should we make each layer have a blend mode (?)

    def process(
        self,
        inputs: dict[str, Union['np.ndarray', 'moderngl.Texture']],
        **kwargs
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Process the `inputs` provided and obtain a result
        as the output.
        """
        base = self.layers[0].process(
            inputs = inputs,
            **kwargs
        )

        for layer, mode in zip(self.layers[1:], self.blend_modes[1:]):
            top = layer.process(
                inputs = inputs,
                **kwargs
            )
            base = self._blend(base, top, mode)

        return base
    
    @abstractmethod
    def _blend(
        self,
        base: Union['np.ndarray', 'moderngl.Texture'],
        top: Union['np.ndarray', 'moderngl.Texture'],
        # TODO: What is the mode (?)
        mode = str
    ) -> Union['np.ndarray', 'moderngl.Texture']:
        """
        Blend the `base` and `top` inputs provided, using
        the also given `mode`, to obtain an output.
        """
        pass